/**
 * JASMINe
 * Copyright (C) 2009 Bull S.A.S.
 * Contact: jasmine@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: MonitoringService.java 6517 2010-05-25 17:25:29Z alitokmen $
 * --------------------------------------------------------------------------
 */
package org.ow2.jasmine.monitoring.eos.monitoring.service;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.ow2.jasmine.event.beans.JasmineEventEB;
import org.ow2.jasmine.monitoring.eos.common.JasmineEventManager;
import org.ow2.jasmine.monitoring.eos.common.jms.JMSConnection;
import org.ow2.jasmine.monitoring.eos.common.jms.event.JasmineMessageEvent;
import org.ow2.jasmine.monitoring.eos.monitoring.DataBaseEventExtractor;
import org.ow2.jasmine.monitoring.eos.monitoring.EventData;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Services provided by the monitoring module.
 *
 * @author Guillaume Renault
 *
 */
public class MonitoringService implements Serializable {

    /**
     * serialVersionUID
     */
    private static final long serialVersionUID = 8402643661543808499L;

    /**
     * Logger.
     */
    private Log logger = LogFactory.getLog(this.getClass());

    private static final String noValue = "noValue";

    /**
     * Timer for disconnecting the JMS connection.
     */
    private transient Timer timer = null;

    /**
     * Time at which to disconnect the JMS topic connection.
     */
    private long disconnectTime = 0;

    /**
     * JMS check interval.
     */
    private static final long JMS_CHECK_INTERVAL = 1000;

    /**
     * JMS timeout.
     */
    private static final long JMS_TIMEOUT = 60 * 1000;

    /**
     * JMS topic used for synchronization.
     */
    private static final String JMS_TOPIC = "jasmine";

    private JMSConnection jmsConnection = null;

    private JasmineEventManager eventManager = null;

    public MonitoringService() {
        this.eventManager = new JasmineEventManager();
    }

    /**
     * Load an xml file to configure graphs.
     *
     * @param file the xml file to load
     * @return a string representing the xml file, without the header.
     */
    public String loadFile(final String file) throws Exception {
        String result = "";
        FileReader f = null;
        BufferedReader buffer = null;

        try {
            String line = "";

            f = new FileReader(file);

            buffer = new BufferedReader(f);

            boolean write = false;

            while ((line = buffer.readLine()) != null) {
                if (line.contains("<monitor")) {
                    write = true;
                }
                if (write) {
                    result += line + "\n";
                }
            }

            return result;
        } finally {
            if (buffer != null) {
                try {
                    buffer.close();
                    buffer = null;
                    System.gc();
                } catch (Exception e) {
                    logger.error("Failed closing buffer", e);
                }
            }

            if (f != null) {
                try {
                    f.close();
                    f = null;
                    System.gc();
                } catch (Exception e) {
                    logger.error("Failed closing file", e);
                }
            }
        }
    }

    /**
     * Load a file to replay.
     *
     * @param fileName the file to replay.
     * @param filter the header fields to get.
     * @return a list of data providers.
     * @throws Exception
     */
    public List<List<EventData>> loadLogFile(final String fileName, final String filter) throws Exception {

        //data which will be looked for in the log file
        String[] toKeep = filter.split("bigSeparator");

        ArrayList<Filter> elements = new ArrayList<Filter>();
        for (int i = 0; i < toKeep.length; i++) {
            elements.add(new Filter(toKeep[i]));
        }

        // the first column is the time, always returned !!

        int[] columns = new int[elements.size()];

        List<List<EventData>> result = new ArrayList<List<EventData>>();
        List<EventData>[] lists = new ArrayList[columns.length];

        // initialize arrays
        for (int i0 = 0; i0 < lists.length; i0++) {
            lists[i0] = new ArrayList<EventData>();
            // the first element is the column name
            lists[i0].add(new EventData(elements.get(i0).header));
        }

        FileReader f = null;
        BufferedReader buffer = null;

        Pattern pattern = null;
        Matcher matcher = null;

        String header = "";
        String line = "";
        f = new FileReader(fileName);
        buffer = new BufferedReader(f);

        // get the header of the log file
        header = buffer.readLine();

        // set an array of the column numbers to keep the values
        String[] headerArray = header.split(";");

        boolean dateIsFirst;
        if (headerArray[0].equals("date")){
            dateIsFirst = true;
        }
        else {
            dateIsFirst = false;
        }
        // determine the index of the csv line for each element (serie) for supporting multi-attributes CSV files
        // init with a single attrib format (field 7)
        for (int i1 = 0; i1 < elements.size(); i1++) {
            columns[i1] = 6;
            for (int i2 = 0; i2 < headerArray.length; i2++) {
                if (elements.get(i1).header.compareTo(headerArray[i2]) == 0) {
                    columns[i1] = i2;
                    break;
                }
            }
        }

        Map<Filter, Pattern> patterns = new HashMap<Filter, Pattern>();

        // compile all the patterns
        for (Filter fi : elements) {
            patterns.put(fi, Pattern.compile(fi.pattern));
        }


        while ((line = buffer.readLine()) != null) {
            // split the line just once
            String[] split = line.split(";");

            for (int i3 = 0; i3 < columns.length; i3++) {

                if (line.toString().contains(lists[i3].get(0).getType()) == true) {
                    matcher = patterns.get(elements.get(i3)).matcher(line.toString());
                    if (elements.get(i3).pattern.equals(noValue) || matcher.matches()) {
                        EventData ed;
                        if (dateIsFirst == true) {
                            ed = new EventData(split[0], split[columns[i3]]);
                        } else {
                            ed = new EventData(split[1], split[columns[i3]]);
                        }
                        ed.setId(elements.get(i3).id);
                        lists[i3].add(ed);
                    }
                }
            }
        }

        // add the lists to the result
        for (List<EventData> l : lists) {
            result.add(l);
        }

        return result;
    }

    /**
     * Loading data from DataBase
     *
     * @param args the arguments to access to database
     * @param filter the fields to get.
     * @return
     * @throws Exception
     */

    public List<List<EventData>> loadFromDataBase(final String args, final String filter) throws Exception {

        Matcher matcher = null;
        EventData ed;

        //split the arguments to get the dates
        String[] datesSelected = args.split("separator");

        Calendar minDate = Calendar.getInstance();
        Calendar maxDate = Calendar.getInstance();
        minDate.set(Integer.parseInt(datesSelected[0]),
            Integer.parseInt(datesSelected[1]),
            Integer.parseInt(datesSelected[2]),
            Integer.parseInt(datesSelected[3]),
            Integer.parseInt(datesSelected[4]));
        maxDate.set(Integer.parseInt(datesSelected[5]),
            Integer.parseInt(datesSelected[6]),
            Integer.parseInt(datesSelected[7]),
            Integer.parseInt(datesSelected[8]),
            Integer.parseInt(datesSelected[9]));
        if (minDate.getTimeInMillis() > maxDate.getTimeInMillis()){
            maxDate.set(Integer.parseInt(datesSelected[0]),
                Integer.parseInt(datesSelected[1]),
                Integer.parseInt(datesSelected[2]),
                Integer.parseInt(datesSelected[3]),
                Integer.parseInt(datesSelected[4]));
            minDate.set(Integer.parseInt(datesSelected[5]),
                Integer.parseInt(datesSelected[6]),
                Integer.parseInt(datesSelected[7]),
                Integer.parseInt(datesSelected[8]),
                Integer.parseInt(datesSelected[9]));
        }

        String[] toKeep = filter.split("bigSeparator");

        ArrayList<Filter> elements = new ArrayList<Filter>();
        for (int i = 0; i < toKeep.length; i++) {
            elements.add(new Filter(toKeep[i]));
        }

        int[] columns = new int[elements.size()];

        List<List<EventData>> result = new ArrayList<List<EventData>>();
        List<EventData>[] lists = new ArrayList[columns.length];

        // initialize arrays
        for (int i0 = 0; i0 < lists.length; i0++) {
            lists[i0] = new ArrayList<EventData>();
            // the first element is the column name
            lists[i0].add(new EventData(elements.get(i0).header));
        }

        Map<Filter, Pattern> patterns = new HashMap<Filter, Pattern>();
        // compile all the patterns
        for (Filter fi : elements) {
            patterns.put(fi, Pattern.compile(fi.pattern));
        }

        //initialize the extractor
        DataBaseEventExtractor dataBaseEventExtractor = new DataBaseEventExtractor();
        JasmineEventEB[] events = null;

        //create strings which represents the probes and the servers
        String probe = "";
        String server = "";
        for (int j = 0; j < elements.size(); j++){
            if (!probe.contains(elements.get(j).header) && !elements.get(j).header.equals(noValue)) {
                if (probe.equals("")) {
                    probe = elements.get(j).header;
                } else {
                    probe = elements.get(j).header + "separator" + probe;
                }
            }
            if (!server.contains(elements.get(j).server) && !elements.get(j).server.equals(noValue)) {
                if (server.equals("")) {
                    server = elements.get(j).server;
                } else {
                    server = elements.get(j).server + "separator" + server;
                }
            }
        }
        String[] arguments = {
                "-Context.INITIAL_CONTEXT_FACTORY", "org.ow2.carol.jndi.spi.MultiOrbInitialContextFactory",
                "-Wrapper.LOOKUP_BEAN", "db-ejb/event",
                //FIXME : the underlying context (JOnAS one) must be used and thus url must not be passed here
                //"-Context.PROVIDER_URL", "rmi://localhost:1099",
                "-domain", "",
                "-server", server,
                "-source", "",
                "-probe", probe,
                "-starttimestamp", String.valueOf(minDate.getTimeInMillis()),
                "-endtimestamp", String.valueOf(maxDate.getTimeInMillis()),
                "-orderby", "timestamp",
                "-limit", "300000"
                };
        //connect to EJB and get the events from database
        events = dataBaseEventExtractor.getEvents(arguments);

        // fill in arrays with the data collected from the database
        // according to the filter related to the graphs
        if (events != null) {
            for (JasmineEventEB eventEB : events) {
                for (int i = 0; i < lists.length; i++) {
                    if (eventEB.getProbe().contains(lists[i].get(0).getType())){
                        matcher = patterns.get(elements.get(i)).matcher(eventEB.toCSV());
                        if (elements.get(i).pattern.equals(noValue) || matcher.matches()) {
                            // convert data to the right format for flex
                            SimpleDateFormat format = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss" );
                            String dateConverted = format.format(eventEB.getTimestamp());

                            ed = new EventData(dateConverted, eventEB.getValue().toString());
                            ed.setId(elements.get(i).id);
                            lists[i].add(ed);
                        }
                    }
                }
            }
        }

        // add the lists to the result
        for (List<EventData> l : lists) {
            result.add(l);
        }

        return result;
    }

    /**
     * Real Time monitoring.
     *
     * @param filter the fields to get.
     * @return
     * @throws Exception
     */
    public List<List<EventData>> startMonitoring(final String filter) throws Exception {
        if (this.jmsConnection == null) {
            this.connectTopic();
        }

        this.resetTimer();

        // ////////////////////////////
        // Initialize the structure //
        // ////////////////////////////

        String[] toKeep = filter.split("bigSeparator");

        ArrayList<Filter> elements = new ArrayList<Filter>();
        for (int i = 0; i < toKeep.length; i++) {
            elements.add(new Filter(toKeep[i]));
        }

        // the first column is the time, always returned !!
        int[] columns = new int[elements.size()];

        List<List<EventData>> result = new ArrayList<List<EventData>>();
        List<EventData>[] lists = new ArrayList[columns.length];

        // initialize arrays
        for (int i0 = 0; i0 < lists.length; i0++) {
            lists[i0] = new ArrayList<EventData>();
            // the first element is the column name
            lists[i0].add(new EventData(elements.get(i0).header));
        }

        // /////////////////////////
        // Analyse JasmineEvents //
        // /////////////////////////

        Pattern pattern = null;
        Matcher matcher = null;

        String probe = "";
        String[] p = null;

        List<JasmineMessageEvent> events = this.eventManager.getAll();
        Iterator<JasmineMessageEvent> eventsIter = events.iterator();

        while (eventsIter.hasNext()) {
            JasmineMessageEvent elem = eventsIter.next();
            // get the probe name
            p = elem.getProbe().split(":");
            probe = p[p.length - 1];

            for (int i = 0; i < lists.length; i++) {
                if (lists[i].get(0).getType().compareTo(probe) == 0) {
                    pattern = Pattern.compile(elements.get(i).pattern);
                    matcher = pattern.matcher(elem.getCsvFormat());
                    if (elements.get(i).pattern.equals(noValue) || matcher.matches()){
                        EventData ed;
                        ed = new EventData(elem.getFormattedTimeStamp(), elem.getValue());
                        ed.setId(elements.get(i).id);
                        lists[i].add(ed);
                    }
                }
            }
        }

        // add the lists to the result
        for (List<EventData> l : lists) {
            result.add(l);
        }

        return result;
    }

    public void stopMonitoring() throws Exception {
        this.disconnectTopic();
    }

    /*
     * ########################################################################
     *
     *
     * Private functions
     *
     *
     * ######################################################################
     */

    /**
     * Connect to the JMS topic.
     */
    private void connectTopic() throws Exception {
        logger.debug("Connecting to the JMS topic...");

        synchronized (this) {
            if (this.jmsConnection == null) {
                this.jmsConnection = new JMSConnection(this.eventManager);
                this.jmsConnection.connect(MonitoringService.JMS_TOPIC);
            }

            this.resetTimer();
        }

        logger.debug("Connected to the JMS topic");
    }

    /**
     * Disconnect from the JMS topic.
     */
    private void disconnectTopic() throws Exception {
        logger.debug("Disconnecting from the JMS topic...");

        synchronized (this) {
            if (this.jmsConnection != null) {
                this.jmsConnection.disconnect();
                this.jmsConnection = null;
            }

            this.disconnectTime = 0;

            if (this.timer != null) {
                this.timer.cancel();
                this.timer.purge();
                this.timer = null;
            }
        }

        logger.debug("Disconnected from the JMS topic");
    }

    /**
     * Reset the timer.
     */
    private void resetTimer() {
        synchronized (this) {
            this.disconnectTime = System.currentTimeMillis() + MonitoringService.JMS_TIMEOUT;

            if (this.timer == null) {
                this.timer = new Timer();
                this.timer.schedule(new DisconnectTask(), MonitoringService.JMS_CHECK_INTERVAL,
                    MonitoringService.JMS_CHECK_INTERVAL);
            }
        }
    }

    /**
     * Implementation of TimerTask.
     *
     * @author Guillaume Renault
     */
    class DisconnectTask extends TimerTask {

        @Override
        public void run() {
            try {
                if (System.currentTimeMillis() > disconnectTime) {
                    disconnectTopic();
                }
            } catch (Throwable t) {
                logger.error("Failed disconnecting from JMS topic", t);
            }
        }
    }

    @SuppressWarnings("unused")
    private void readObject(final ObjectInputStream in) throws IOException, ClassNotFoundException {
        // deserialize everything except for the transient fields
        in.defaultReadObject();

        if (this.disconnectTime > 0) {
            // relaunch JMS timer
            this.resetTimer();
        }
    }

    class Filter {

        private String header;
        private String id;
        private String pattern;
        private String server;

        public Filter(final String properties){
            String[] tabString = properties.split("separator");
            this.header = tabString[0];
            this.id = tabString[1];
            this.pattern = tabString[2];
            this.server = tabString[3];
        }

        public String getHeader() {
            return header;
        }

        public String getId() {
            return id;
        }

        public String getPattern() {
            return pattern;
        }

        public String getServer() {
            return server;
        }

    }

}
